#!/usr/bin/python

import pprint
import curses
import collections
import copy
import xml.dom.minidom as minidom
import urllib2
from SOAPpy import *

UPnPActions = ['SetConnectionType',
	   'GetConnectionTypeInfo',
	   'ConfigureConnection',
	   'RequestConnection',
	   'RequestTermination',
	   'ForceTermination',
	   'SetAutoDisconnectTime',
	   'SetIdleDisconnectTime',
	   'SetWarnDisconnectDelay',
	   'GetStatusInfo',
	   'GetLinkLayerMaxBitRates',
	   'GetPPPEncryptionProtocol',
	   'GetPPPCompressionProtocol',
	   'GetPPPAuthenticationProtocol',
	   'GetUserName',
	   'GetPassword',
	   'GetAutoDisconnectTime',
	   'GetIdleDisconnectTime',
	   'GetWarnDisconnectDelay',
	   'GetNATRSIPStatus',
	   'GetGenericPortMappingEntry',
	   'GetSpecificPortMappingEntry',
	   'AddPortMapping',
	   'DeletePortMapping',
	   'GetExternalIPAddress']

class uCTUI():


	def __init__(self, type, selectedIP, selectedIPXML, upnpList, serviceList, mapList, wrappedScr):

		self.type = type
		self.scan = ''
		self.mapList = mapList
		self.upnpList = upnpList
		self.selectedIPXML = selectedIPXML
		self.serviceList = serviceList
		self.messagesQueue = collections.deque()
		self.positiveQueue = collections.deque()
		self.cursorPos = 0
		self.selectedIP = selectedIP
		self.qSize = 0
		self.mainScreen = wrappedScr
		self.mainMax = self.mainScreen.getmaxyx()
		if self.mainMax[0] < 40:
			curses.endwin()
			print "Terminal too small"
			sys.exit(2)
		if self.mainMax[1] < 60:
			curses.endwin()
			print "Terminal too small"
			sys.exit(2)
		self.mainScreen.clear()
		self.mainScreen.subwin(self.mainMax[0], self.mainMax[1], 0, 0)
		self.messagesLimit = self.mainMax[1]
		self.mainScreen.box()
		self.mainScreen.refresh()

	def draw(self):
		start = 0
		if self.type == 1:
			self.positiveWindow = curses.newwin(20,20,1,1)
			self.positiveWindow.box()
			self.positiveWindow.addstr(0,1,"Positive IP's",curses.A_REVERSE)
			self.positiveWindow.refresh()

			self.detailWindow = curses.newwin(20,self.mainMax[1]-25,1,22)
			self.detailWindow.box()
			if len(self.upnpList) > 0:
				dselectedIP = self.upnpList[self.selectedIP].split('|')[0]
				dselectedService = self.upnpList[self.selectedIP].split('|')[7].split(':')[3]
				dselectedTiming = self.upnpList[self.selectedIP].split('|')[8].split('.')[0]
			else:
				dselectedIP = ''
				dselectedService = ''
				dselectedTiming = '-'
			self.detailWindow.addstr(0,1,"UPnP Details -  |Selected IP: "+dselectedIP+" |Total positives: "+str(len(self.upnpList))+" |Service: "+dselectedService+" |Timing: "+dselectedTiming+"ms",curses.A_REVERSE)
			dcount = 1
			for k in self.selectedIPXML:
				if dcount < 20: self.detailWindow.addstr(dcount,1,str(k)+": "+str(self.selectedIPXML[k]))
                        	dcount = dcount + 1
                        
                        
                        if len(dselectedIP) > 0:
                        	dcount = dcount + 1
                        	displayList = []
                        	for serv in self.serviceList[dselectedIP].split(","):
                        		displayService = str(serv).split("|")[0]
                        		if displayService not in displayList:
                        			if dcount < 20: self.detailWindow.addstr(dcount,1,displayService)
                        			dcount = dcount + 1
                        			displayList.append(displayService)
	                self.detailWindow.refresh()

			

			self.messagesWindow = curses.newwin(self.mainMax[0]-22, self.mainMax[1]-4, 21, 1)
			self.messagesWindow.box()
			self.messagesWindow.addstr(0,1,"Messages",curses.A_REVERSE)
			self.messagesWindow.refresh()
			count = 1
			if self.selectedIP >= range(19,254,19)[0]:
				start = self.selectedIP
				end = self.selectedIP+19
			elif self.selectedIP >= range(19,254,19)[1]:
				start = self.selectedIP
				end = self.selectedIP+19

			for positive in self.upnpList:
				if count < start+1:
					count += 1
					continue
				if self.selectedIP < start+19 and start > 18 and count < end:
					self.positiveWindow.addstr(count-start,1,positive.split('|')[0])
	                        	count = count + 1
#	                        if self.selectedIP == count-1:
#					self.positiveWindow.addstr(count-start,1,str(count-start)+"+"+positive.split('|')[0],curses.A_REVERSE)
#					count = count + 1
#					continue
				if count < 19 and start == 0:
					self.positiveWindow.addstr(count,1,positive.split('|')[0])
	                        	count = count + 1
	                self.positiveWindow.refresh()

		else:
		
			self.mappingWindow = curses.newwin(20,47,1,1)
			self.mappingWindow.box()
			self.mappingWindow.addstr(0,1,"Mappings",curses.A_REVERSE)
			mcount = 1
			for map in self.mapList:
				if mcount < 20: 
					self.mappingWindow.addstr(mcount,1,str(map))
					mcount = mcount + 1
			self.mappingWindow.refresh()

			self.detailWindow = curses.newwin(20,self.mainMax[1]-52,1,48)
			self.detailWindow.box()
			self.detailWindow.addstr(0,1,"UPnP Details",curses.A_REVERSE)
			dcount = 1
			for k in self.selectedIPXML:
				if dcount < 20: self.detailWindow.addstr(dcount,1,str(k)+": "+str(self.selectedIPXML[k]))
                        	dcount = dcount + 1
                        dcount = dcount + 2
                        for s in self.serviceList:
                        	if dcount < 20: self.detailWindow.addstr(dcount,1,str(s))
                        	dcount = dcount + 1
	                self.detailWindow.refresh()

			it = self.mainMax[0]-22
			self.messagesWindow = curses.newwin(self.mainMax[0]-22, self.mainMax[1]-4, 21, 1)
			self.messagesWindow.box()
			self.messagesWindow.addstr(0,1,"Messages",curses.A_REVERSE)
			self.messagesWindow.refresh()
			count = 1
			for map in self.mapList:
				if count < 20: self.mappingWindow.addstr(count,1,str(count)+"-"+map)
				count = count + 1
				
		qCopy = copy.deepcopy(self.messagesQueue)
		count = 1
		for message in qCopy:
			self.messagesWindow.addstr(count,1,str(message))
			count = count + 1
		self.messagesWindow.refresh()




			
	def drawExecute(self):
		
		
		actionNumber = {}
		serviceNumber = {}
		actionList = []
		argumentDict = {}
		stateVars = {}
		inputArgs = {}
		splitService = ''
		if len(self.upnpList) == 0:
			return False
		self.executeWindow = curses.newwin(40,self.mainMax[1]-25,3,5)
	        self.executeWindow.box()
		self.executeWindow.refresh()
		count = 0
		thisIP = self.upnpList[self.selectedIP].split('|')[0]
		sl = []
		for service in self.serviceList[thisIP].split(","):
			strCount = str(count)
			splitService = service.split("|")
			if splitService[0] == '':
				continue
			if splitService[0] not in sl:
				self.executeWindow.addstr(count+1,1,strCount+"-"+str(splitService[0]))
			sl.append(splitService[0])
			serviceNumber[strCount] = str(splitService[0])
			count = count + 1
		serviceSelection = self.executeWindow.getstr()
		self.executeWindow.clear()
		self.executeWindow.refresh()
		self.executeWindow.box()
		thisXML = self.serviceList[thisIP].split(",")[int(serviceSelection)].split('|')[1]
		thisType = self.serviceList[thisIP].split(",")[int(serviceSelection)].split('|')[0]
		thisControl = self.serviceList[thisIP].split(",")[int(serviceSelection)].split('|')[2]
		try:
			nowIP = self.upnpList[self.selectedIP].split('|')[0]
			nowPort = self.upnpList[self.selectedIP].split('|')[2]
			nowXML = thisXML
#			nowXML = self.upnpList[self.selectedIP].split('|')[6]
			scpd = "http://"+nowIP+":"+nowPort+nowXML
	                request = urllib2.Request(scpd, None)
	                response = urllib2.urlopen(request)
	                headers = response.info()
	                xmlData = response.read()
                except Exception, err:
            		return False
		xmlRoot = minidom.parseString(xmlData)
		for stateVar in xmlRoot.getElementsByTagName('stateVariable'):
			stateVarName = stateVar.getElementsByTagName('name')[0].childNodes[0].data
			stateVarType = stateVar.getElementsByTagName('dataType')[0].childNodes[0].data
			stateVars[str(stateVarName)] = str(stateVarType)

		for action in xmlRoot.getElementsByTagName('action'):
			actionName = action.getElementsByTagName('name')[0].childNodes[0].data
			actionList.append(str(actionName))
			argumentDict[actionName] = ''
			for argument in action.getElementsByTagName('argument'):
				lenAction = len(actionList)
				nowArg = argument.getElementsByTagName('name')[0].childNodes[0].data
				
				relatedVar = argument.getElementsByTagName('relatedStateVariable')[0].childNodes[0].data
				direction = argument.getElementsByTagName('direction')[0].childNodes[0].data
				if argumentDict[actionName] == '':
					argumentDict[actionName] = nowArg+":"+stateVars[relatedVar]+":"+direction
				else:
					argumentDict[actionName] = argumentDict[actionName]+"|"+nowArg+":"+stateVars[relatedVar]+":"+direction

		count = 1
		for action in actionList:
			strCount = str(count)
			self.executeWindow.addstr(count,1,strCount+"-"+str(action))
			actionNumber[strCount] = str(action)
			count = count + 1
			
#		for action in UPnPActions:
#			strCount = str(count)
#			if action in actionList: continue
#			self.executeWindow.addstr(count,1,strCount+"-"+str(action),curses.A_BOLD)
#			actionNumber[strCount] = str(action)
#			count = count + 1
		self.executeWindow.refresh()

		selection = self.executeWindow.getstr()

		if not selection.isdigit() or int(selection) > count:
			self.mainScreen.clear()
			self.mainScreen.refresh()
			self.printMessage("Error in selection")
			return False
		try:
			nowAction = actionNumber.get(selection)
			
			if argumentDict.has_key(nowAction): 
				nowArgs = argumentDict[nowAction].split('|')
			else:
				nowArgs = ''
			for nowArg in nowArgs:
				nowArg = nowArg.split(':')
				if nowArg[2] == 'out': continue
				self.executeWindow.clear()
				self.executeWindow.refresh()
				curses.echo()
				self.executeWindow.addstr(1,1,"Insert("+str(nowArg[1])+") "+str(nowArg[0])+":")
				input = self.executeWindow.getstr()
				curses.noecho()
				inputArgs[str(nowArg[0])] = str(input)+"|"+str(nowArg[1])
				self.executeWindow.refresh()
		except Exception, err:
			self.printMessage("Error in selection "+str(err))
			return False 
		sendParms = self.upnpList[self.selectedIP].split('|')
		if selection == '99':
			soapRequest = self.sendSoap(thisControl, sendParms[0], sendParms[2], thisType, "GetPPPAuthenticationProtocol", inputArgs)
		else:
			soapRequest = self.sendSoap(thisControl, sendParms[0], sendParms[2], thisType, actionNumber.get(selection), inputArgs)
		newXML = ''
		if not soapRequest or 'faultString' in soapRequest:
			self.printMessage("Error in UPnP Request")
			return False
		try:
			newXML = '\n'.join(soapRequest)

			xmlResponse = minidom.parseString(newXML)
			
			ecount = 1
			self.executeWindow.clear()
			self.executeWindow.refresh()
			nowArgs = ''
			for r in xmlResponse.getElementsByTagName('s:Body'):
				
				for response in r.childNodes:
					for answer in response.childNodes:
						if answer.firstChild is not None:
							self.executeWindow.addstr(ecount,1,answer.tagName+": "+answer.firstChild.nodeValue)
						ecount = ecount + 1
				
			self.executeWindow.addstr(ecount+5,1,"Press any key to continue")
			self.executeWindow.refresh()
			self.executeWindow.getch()
		except Exception, err:
			self.printMessage("Exception: "+str(err))
			return False
			
		del self.executeWindow
		self.mainScreen.clear()
		self.mainScreen.refresh()

		


	def sendSoap(self, control, ip, port, device, action, inputArgs = ''):
		body = []
		bodyStr = '<ns1:'+action+' xmlns:ns1="'+device+'" SOAP-ENC:root="1">\n'
		if len(inputArgs) > 0:
			for k,v in inputArgs.iteritems():
				splitv = v.split('|')
				try:
					body.append("<"+k+" xsi:type=\"xsd:"+splitv[1]+"\">"+splitv[0]+"</"+k+">")
				except Exception, err:
					self.printMessage("Exception "+str(v))
			for b in body:
				bodyStr = bodyStr+str(b)+"\n"
		bodyStr = bodyStr+"</ns1:"+action+">"
		payload = """<?xml version="1.0" encoding="UTF-8"?>
<SOAP-ENV:Envelope
  SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"
  xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
  xmlns:xsi="http://www.w3.org/1999/XMLSchema-instance"
  xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"
  xmlns:xsd="http://www.w3.org/1999/XMLSchema"
>
<SOAP-ENV:Body>"""+bodyStr+"""
</SOAP-ENV:Body>
</SOAP-ENV:Envelope>
\r\n\r\n
"""

		head = """POST """+control+""" HTTP/1.0\r
Host: """+ip+"""\r
User-agent: Umap v0.8 UPnP/1.1 None\r
Content-type: text/xml; charset="UTF-8"\r
Content-length: """+str(len(payload))+"""\r
SOAPAction: """+'"'+device+'#'+action+'"'+"""\r
\r\n"""


		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		try:
			sock.connect((ip, int(port)))
		except Exception, err:
			self.printMessage("Error connecting for SOAP request")
			return False
		sock.send(head+payload)
		soapRcvStr = ''
		while 1:
			try:
				soapRcv = sock.recv(2048)
			except Exception as err:
				if 'timed out' in err:
					continue
				else:
					self.printMessage("Error receiving "+str(err))
					return False
			if not soapRcv: break
			soapRcvStr += soapRcv
		sock.close
		f = 0
		response = []
		for s in soapRcvStr.split('\n'):
			if '<?xml ' in s or '<s:Envelope' in s: f = 1
			if f == 1: 
				response.append(s)

				
		return response
		
	def printMessage(self, message):

		queueSize = len(self.messagesQueue)
		self.qSize = queueSize
		total = self.mainMax[0]-24
		min = total+1
		max = total-1
		count = 1
		if queueSize == total:
			popleft = self.messagesQueue.popleft()
			self.messagesQueue.append(message)
		else:
			self.messagesQueue.append(message)

	
	def printInfo(self, out):

		count = 1
		for k in out:
			if count < 20: self.detailWindow.addstr(count,1,str(k)+": "+str(out[k]))
                        count = count + 1
                self.detailWindow.refresh()

	def cursorUp(self):

		self.selectedIPXML = {}
		if self.cursorPos == 1 or self.selectedIP <= 0: 
			return self.cursorPos
		else:
			self.cursorPos = self.cursorPos - 1
			self.selectedIP = self.selectedIP - 1
			return self.cursorPos
		

	def cursorDown(self):

		self.selectedIPXML = {}
		if len(self.upnpList) == self.cursorPos or self.selectedIP == len(self.upnpList)-1:
			return self.cursorPos
		else:
			self.cursorPos = self.cursorPos + 1
			self.selectedIP = self.selectedIP + 1
			return self.cursorPos

	def allRefresh(self):

		self.mainScreen.refresh()
		self.positiveWindow.refresh()
		self.detailWindow.refresh()
		self.messagesWindow.refresh()

	def exitCurse(self):

		curses.endwin()
