#!/usr/bin/python


import cmd
import readline
import collections
import copy
import xml.dom.minidom as minidom
import urllib2
import iplib
import upnpscan, uscan
import threading
from SOAPpy import *


class uTUI(cmd.Cmd):
	
	def __init__(self, ip, queue, upnpList, verbose, serviceList, portList):
		cmd.Cmd.__init__(self)
		self.intro = "Umap v0.8"
		self.prompt = "umap>"
		self.env = {}
		self.argList = ['IP',
				'Service',
				'Verbose',
				'MaxThreads']
		for a in self.argList:
			self.env[a.lower()] = ''
		self.env['maxthreads'] = '64'
		self.portList = portList
		self.mapList = []
		self.serviceList = serviceList
		self.selectedIP = ''
		self.selectedService = ''
		self.queue = queue
		self.maxThreads = 16
		self.upnpList = upnpList
		self.verbose = verbose
		self.selectedIPXML = {}
		self.ip = ip 

        def do_scan(self, args):
		"""Scan IP/IPCIDR/IPCSV for common XML description files"""
		if not self.env.has_key('ip'):
			print("Need IP for scanning "+self.env['ip'])
			return
		if self.env.has_key('maxthreads'):
			maxThreads = self.env['maxthreads']

		if '/' in self.env['ip']:
			CIDR = iplib.CIDR(self.env['ip'])
			for i in CIDR.get_all_valid_ip():
				self.queue.put(i.get())
		elif ',' in self.env['ip']:
			for i in self.env['ip'].split(','):
			    self.queue.put(i)
		else:
			self.queue.put(self.env['ip'])

		if self.env.has_key('verbose') or self.verbose == 1:
			self.verbose = self.env['verbose']
		else:
			self.verbose = 0
		totalCount = 0
		while True:
			if self.queue.empty(): break
			if (threading.activeCount() < int(maxThreads)):
                        	current_ip = self.queue.get()
                                self.upnpScan = upnpscan.UPNPScan(current_ip, self.upnpList, '', self.selectedIPXML, self.serviceList, self.verbose)
                                totalCount += 1
                                print("Scanning("+str(threading.activeCount())+","+str(totalCount)+"): "+current_ip)
                                self.upnpScan.setDaemon(True)
                                self.upnpScan.start()
                
                	

	def do_scanLAN(self, args):
		"""Scan internal LAN on positive IP"""
		if len(self.upnpList) == 0:
			print("Need a positive match to scan the LAN")
			return False
		if self.env.has_key('verbose') or self.verbose == 1:
                        self.verbose = self.env['verbose']
                else:
                        self.verbose = 0

		splitUPnP = self.upnpList[self.selectedIP].split('|')
		base = splitUPnP[9]
		ipGuess = base.split('://')[1].split(':')[0]
		(a,b,c,d) = ipGuess.split('.')

                d = int(d)
		d  = 1
		while d < 255:
		        d += 1
		        qIP = a+'.'+b+'.'+c+'.'+str(d)
		        self.queue.put(qIP)

                location = splitUPnP[3]+"|"+splitUPnP[4]+"|"+splitUPnP[5]

		scanned = self.upnpScan.getInfo(location, splitUPnP[0], 1)

		while True:
                    if(threading.activeCount() < self.maxThreads):
                        current_ip = self.queue.get()
                        umap = uscan.Uscan(splitUPnP[0], current_ip, scanned, self.queue, self.verbose, self.portList, self.mapList, '')
                        umap.setDaemon(True)
                        umap.start()

	
	def do_set(self, args):
		"""Set/show available variables and values

	IP (IP of targets to scan. Formats accepted: CIDR, CSV, Single IP)
	Service  (UPnP Service to use)
	MaxThreads  (Maximum amount of threads allowed)
	Verbose"""
		if '=' in args:
		    var = args.split('=')[0].rstrip().lstrip()
		    value = args.split('=')[1].rstrip().lstrip()
		    if '=' in value:
			value = args.split('"')[1]
		    if "'" in value:
			value = args.split("'")[1]
		    self.env[var.lower()] = value
		    return 0

		if args == '' and len(self.env) > 0:
			for v in self.env:
			    print v+"="+self.env[v]
		else:
			print "No vars set\n"
	


	def do_fetchActions(self, args):
		"""Fetch actions advertised by device/service"""
		actionNumber = {}
		serviceNumber = {}
		actionList = []
		argumentDict = {}
		stateVars = {}
		inputArgs = {}
		splitService = ''
		if len(self.upnpList) == 0:
			print "No positive IP's!"
			return False
		if self.selectedIP == '':
			print "Selecting first ip in list, use the command selectIP to select a specific IP"
			self.selectedIP = self.upnpList[0].split("|")[0]
		
		if self.selectedService == '':
			print "Selecting first service in list, use command selectService to select a specific service"
			thisXML = self.serviceList[self.selectedIP].split(",")[0].split('|')[1]
			thisType = self.serviceList[self.selectedIP].split(",")[0].split('|')[0]
			thisControl = self.serviceList[self.selectedIP].split(",")[0].split('|')[2]
		else:
			c = 0
			for i in self.serviceList[self.selectedIP].split(","):
					if i.split('|')[0] == self.selectedService:
						thisXML = self.serviceList[self.selectedIP].split(",")[c].split('|')[1]
						thisType = self.serviceList[self.selectedIP].split(",")[c].split('|')[0]
						thisControl = self.serviceList[self.selectedIP].split(",")[c].split('|')[2]
					c = c + 1
					
		try:
			for u in self.upnpList:
				if u.split('|')[0] == self.selectedIP:
					nowIP = self.upnpList[c].split('|')[0]
					nowPort = self.upnpList[c].split('|')[2]
					nowXML = thisXML
					break
			if not nowIP:
				print "Couldn't find IP in list"
				return False

			
			scpd = "http://"+nowIP+":"+nowPort+nowXML
	                request = urllib2.Request(scpd, None)
	                response = urllib2.urlopen(request)
	                headers = response.info()
	                xmlData = response.read()
	        except Exception, err:
	        	print err
	        	return False
				
		xmlRoot = minidom.parseString(xmlData)
		
		for stateVar in xmlRoot.getElementsByTagName('stateVariable'):
			stateVarName = stateVar.getElementsByTagName('name')[0].childNodes[0].data
			stateVarType = stateVar.getElementsByTagName('dataType')[0].childNodes[0].data
			stateVars[str(stateVarName)] = str(stateVarType)
		
		for action in xmlRoot.getElementsByTagName('action'):
			actionName = action.getElementsByTagName('name')[0].childNodes[0].data
			actionList.append(str(actionName))
			argumentDict[actionName] = ''
			for argument in action.getElementsByTagName('argument'):
				lenAction = len(actionList)
				nowArg = argument.getElementsByTagName('name')[0].childNodes[0].data
				
				relatedVar = argument.getElementsByTagName('relatedStateVariable')[0].childNodes[0].data
				direction = argument.getElementsByTagName('direction')[0].childNodes[0].data
				if argumentDict[actionName] == '':
					argumentDict[actionName] = nowArg+":"+stateVars[relatedVar]+":"+direction
				else:
					argumentDict[actionName] = argumentDict[actionName]+"|"+nowArg+":"+stateVars[relatedVar]+":"+direction
			self.argumentDict = argumentDict

		count = 1
		for action in actionList:
			strCount = str(count)
			print(strCount+"-"+str(action))
			actionNumber[strCount] = str(action)
			count = count + 1

	def do_viewServices(self, args):
		"""View services available"""
		
		if len(self.upnpList) == 0:
			print "No positive results!"
			return False
			
		if self.selectedIP == '' and len(self.upnpList) > 0:
			print "Selecting first ip in list, use the command selectIP to select a specific IP"
			self.selectedIP = self.upnpList[0].split("|")[0]
			
		for service in self.serviceList[self.selectedIP].split(','):
			print service.split('|')[0]
		

	def do_executeAction(self, args):
		"""Execute action"""
		inputArgs = {}
		if self.selectedIP == '':
			print "Need to select IP first"
			return False
		if self.selectedService == '':
			print "Need to select service first"
			return False
			
		c = 0
		for i in self.serviceList[self.selectedIP].split(","):
			if i.split('|')[0] == self.selectedService:
				thisXML = self.serviceList[self.selectedIP].split(",")[c].split('|')[1]
				thisType = self.serviceList[self.selectedIP].split(",")[c].split('|')[0]
				thisControl = self.serviceList[self.selectedIP].split(",")[c].split('|')[2]
			c = c + 1
		
		for u in self.upnpList:
			if u.split('|')[0] == self.selectedIP:
				nowIP = self.upnpList[c].split('|')[0]
				nowPort = self.upnpList[c].split('|')[2]
				nowXML = thisXML
				break
				
		if not nowIP:
			print "Couldn't find IP in UPnP List!"
			return False
			
			
		if self.argumentDict.has_key(args): 
			nowArgs = self.argumentDict[args].split('|')
		else:
			nowArgs = ''
		
		for nowArg in nowArgs:
			nowArg = nowArg.split(':')
			if nowArg[2] == 'out': continue
			print("Insert("+str(nowArg[1])+") "+str(nowArg[0])+":")
			input = raw_input()
			inputArgs[str(nowArg[0])] = str(input)+"|"+str(nowArg[1])


		try:
			response = self.sendSoap(thisControl, self.selectedIP, nowPort, thisType, args, inputArgs)
			newXML = '\n'.join(response)
			xmlResponse = minidom.parseString(newXML)
		
			nowArgs = ''
			for r in xmlResponse.getElementsByTagName('s:Body'):
					
				for response in r.childNodes:
					for answer in response.childNodes:
						if answer.firstChild is not None:
							print(answer.tagName+": "+answer.firstChild.nodeValue)
		except Exception, err:
			print err
			return False


	def do_selectIP(self, args):
		"""Select IP from Positives"""
		self.selectedIP = args
		
	def do_selectService(self, args):
		"""Select UPnP service"""
		self.selectedService = args

	def complete_selectService(self, text, line, begidx, endidx):
		if not text:
			complete = [ f.split('|')[0]
				        for f in self.serviceList[self.selectedIP].split(',')
				        if True
				        ]
		else:
			complete = [ f.split('|')[0]
					for f in self.serviceList[self.selectedIP].split(',')
					if f.split('|')[0].startswith(text)
					]
		return complete

	def complete_executeAction(self, text, line, begidx, endidx):		
		if not text:
			complete = [ f
				        for f in self.argumentDict
				        if True
				        ]
		else:
			complete = [ f
					for f in self.argumentDict
					if f.startswith(text)
					]
		return complete

		
	def complete_selectIP(self, text, line, begidx, endidx):
		if not text:
			complete = [ f.split('|')[0]
				        for f in self.upnpList
				        if True
					]
		else:
			complete = [ f.split('|')[0]
					for f in self.upnpList
					if f.startswith(text)
					]
		return complete

	def do_showPositives(self, args):
		"""Show positive IP's found from scan"""
		self.prompt = ""
		c = 0
		for i in self.upnpList:
			print "["+str(c)+"] "+i.split('|')[0]
			c += 1
		self.prompt = "umap>"

	def do_showInfo(self, args):
		"""Show advertised modem information"""
		print args
		
	def sendSoap(self, control, ip, port, device, action, inputArgs = ''):
		body = []
		bodyStr = '<ns1:'+action+' xmlns:ns1="'+device+'" SOAP-ENC:root="1">\n'
		if len(inputArgs) > 0:
			for k,v in inputArgs.iteritems():
				splitv = v.split('|')
				try:
					body.append("<"+k+" xsi:type=\"xsd:"+splitv[1]+"\">"+splitv[0]+"</"+k+">")
				except Exception, err:
					print("Exception "+str(v))
			for b in body:
				bodyStr = bodyStr+str(b)+"\n"
		bodyStr = bodyStr+"</ns1:"+action+">"
		payload = """<?xml version="1.0" encoding="UTF-8"?>
<SOAP-ENV:Envelope
  SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"
  xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
  xmlns:xsi="http://www.w3.org/1999/XMLSchema-instance"
  xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"
  xmlns:xsd="http://www.w3.org/1999/XMLSchema"
>
<SOAP-ENV:Body>"""+bodyStr+"""
</SOAP-ENV:Body>
</SOAP-ENV:Envelope>
\r\n\r\n
"""

		head = """POST """+control+""" HTTP/1.0\r
Host: """+ip+"""\r
User-agent: Umap v0.8 UPnP/1.1 None\r
Content-type: text/xml; charset="UTF-8"\r
Content-length: """+str(len(payload))+"""\r
SOAPAction: """+'"'+device+'#'+action+'"'+"""\r
\r\n"""


		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		try:
			sock.connect((ip, int(port)))
		except Exception, err:
			print("Error connecting for SOAP request")
			return False
		sock.send(head+payload)
		soapRcvStr = ''
		while 1:
			try:
				soapRcv = sock.recv(4096)
			except Exception as err:
				if 'timed out' in err:
					continue
				else:
					print("Error receiving "+str(err))
					return False
			if not soapRcv: break
			soapRcvStr += soapRcv
		sock.close
		f = 0
		response = []
		for s in soapRcvStr.split('\n'):
			if '<?xml ' in s or '<s:Envelope' in s: f = 1
			if f == 1: 
				response.append(s)

				
		return response

	
	def do_quit(self, args):
		return -1
		
	def do_exit(self, args):
		return -1

	def emptyline(self):
		pass
	

def guessIP(base, cursesMode):
    # TODO
    # Best guess of IP (10.0.0.0)
    # ipGuess = '10.0.0.1'
    print base
    if base != '/' or base != '':
        ipGuess = base.split('://')[1].split(':')[0]
    else:
        ipGuess = '10.0.0.1'


    (a,b,c,d) = ipGuess.split('.')

    d = int(d)
    originald = d
    d = 1

    if cursesMode:
        self.printMessage("[-] Guessing internal IP's on block "+ipGuess)
    else:
         print("[-] Guessing internal IP's on block "+ipGuess)
    while d < 255:
        d += 1
        ipGuess = a+'.'+b+'.'+c+'.'+str(d)
        queue.put(ipGuess)

